#include <netdef.h>
#include <proto.h>
#include <conn.h>
#include <osport.h>

#ifdef DEBUG
#include <util.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/time.h>

#ifdef _REENTRANT
#include <pthread.h>
#endif

static inline void _c_lock(connection *c)
{
#ifdef _REENTRANT
	if (!c->_locked || c->_locked_thr != pthread_self())
		pthread_mutex_lock(&c->_mutex);
#endif
}

static inline void _c_unlock(connection *c)
{
#ifdef _REENTRANT
	if (!c->_locked || c->_locked_thr != pthread_self())
		pthread_mutex_unlock(&c->_mutex);
#endif
}

int destroy(connection *c){
  disconnect(c);
  free(c); /* this will free up the node */
  return 0;
}

void applyarg(connection *c, int arg){
  if(c == NULL)
    return;
  c->near.junk = arg;
}

connection* get_connection(netnode n, protocol *p, int flags){
  connection *c = NULL;
  netnode *n1 = NULL;

  n1 = getlocalip(n);
  if(n1 == NULL)
    return NULL;

  c = new_connection(*n1, n, p, flags);

  if(c == NULL){
    free(n1);
  }
  return c;
}

/**
 * new_connection returns a new connection object ready to be connected
 */

connection* new_connection(netnode n1, netnode n2, protocol *p, int flags){
  connection *c; 

  c = (connection *)malloc(sizeof(connection));

  if(c == NULL)
    return NULL;
  
  c->status = 0;
  c->num    = 0;
  c->flags = 0;
  applyflag(c, flags);
  c->near   = n1;
  c->far    = n2;
  c->prot   = p;

#ifdef _REENTRANT
  c->_locked = 0;
  pthread_mutex_init(&c->_mutex, NULL);
#endif

  return c;
}

/**
 */
int connectout(connection *c){
  /*  applyflag(c, C); */ 
  int ret = 0;
  if(c == NULL)
    return -1;
  if(checkflag(c, CONN_CONNECTED))
    return -1;

#ifdef _REENTRANT
  _c_lock(c);
#endif

  ret = c->prot->conn(c->near, c->far, c->flags);

  if(ret < 0){
#ifdef _REENTRANT
    _c_unlock(c);
#endif
    return -1;
  }
  applyflag(c, CONN_CONNECTED);
  c->fd = c->near.fd = ret;
#ifdef _REENTRANT
  _c_unlock(c);
#endif

  return ret;
}

/**
 */
int connectin(connection *c, int len){
  int ret = 0;
  if(c == NULL)
    return -1;
  if(checkflag(c, CONN_CONNECTED))
    return -1;

#ifdef _REENTRANT
  _c_lock(c);
#endif
#ifdef DEBUG
  __util_debug_print("connectin()");
#endif

  applyflag(c, CONN_LISTEN);
#ifdef DEBUG
  __util_debug_print("applied flags");
#endif
  applyarg(c, len + 1);
#ifdef DEBUG
  __util_debug_print("applied arguments");
#endif

  ret = c->prot->conn(c->near, c->far, c->flags);
#ifdef DEBUG
  __util_debug_print("got value from connection");
#endif

  if(ret < 0){
#ifdef DEBUG
    __util_debug_print("error!!!!!");
    __util_debug_print("leaving - connectin()");
#endif
#ifdef _REENTRANT
    _c_unlock(c);
#endif
    return -1;
  }
  
  applyflag(c, CONN_CONNECTED);
#ifdef DEBUG
  __util_debug_print("applied connected flag");
#endif
  c->fd = c->near.fd = ret;
#ifdef DEBUG
  __util_debug_print("leaving - connectin()");
#endif
#ifdef _REENTRANT
  _c_unlock(c);
#endif

  return ret;
}

int isdisconnected(connection *c){
  if(checkflag(c, CONN_CONNECTED))
    return 0;
  return 1;
}

connection *accept_new(connection *c){
  netnode *n;
  int fd = 0;
  connection *s;

  if(c == NULL)
    return NULL;
#ifdef _REENTRANT
  _c_lock(c);
#endif

  n = get_node(INADDR_ANY, 0);

  if(n == NULL){
#ifdef DEBUG
    __util_debug_print("Error!! n didn't get allocated!!");
#endif
#ifdef _REENTRANT
    _c_unlock(c);
#endif
    return NULL;
  }
  if(!checkflag(c, CONN_LISTEN)){
#ifdef _REENTRANT
    _c_unlock(c);
#endif
    
    return NULL;
  }

  fd = c->prot->accept(c->near, c->far, c->flags, n);
  
  s = new_connection(c->near, *n, c->prot, c->flags);

  if(s < 0){
#ifdef _REENTRANT
    _c_unlock(c);
#endif
    return NULL;
  }
  applyflag(s, CONN_CHILD);
  
  s->fd = fd;
  s->near.fd = fd;
#ifdef _REENTRANT
  _c_unlock(c);
#endif

  return s;
}

/**
 */
int applyflag(connection *c, int flags){
  if(c == NULL)
    return -1;
  c->flags |= flags;
  return c->flags;
}

int checkflag(connection *c, int flags){
  if(c == NULL)
    return -1;
  return c->flags & flags;
}

int senddata(connection *c, char *buf, int flags, int size){
  int ret = 0;
#ifdef _REENTRANT
  _c_lock(c);
#endif
  if(applyflag(c, flags) >= 0){
    ret = c->prot->send(c->near, c->far, c->flags, buf, size);
  }else{
    ret = -1;
  }
#ifdef _REENTRANT
  _c_unlock(c);
#endif
  return ret;
}

int recvdata(connection *c, char *buf, int flags, int size){
  int ret = 0;
#ifdef _REENTRANT
  _c_lock(c);
#endif

  if(applyflag(c, flags) >= 0)
    ret = c->prot->recv(c->near, c->far, c->flags, buf, size);
  else
    ret = -1;
#ifdef _REENTRANT
  _c_unlock(c);
#endif
  return ret;
}

int disconnect(connection *c){
  int ret = 0;
#ifdef _REENTRANT
  _c_lock(c);
#endif
  if(checkflag(c, CONN_CONNECTED) >= 0)
    ret = c->prot->disc(c->near, c->far);
  else
    ret = -1;
#ifdef _REENTRANT
  _c_unlock(c);
#endif
  return ret;
}

int poll_conn(connection *c, int ms){
  fd_set mask;
  int ret;
  struct timeval tv;

  if(c == NULL)
    return -1;
#ifdef _REENTRANT
  _c_lock(c);
#endif
  
  if(ms >= 0){
    /* tv = (struct timeval *)malloc(sizeof(struct timeval)); */
    tv.tv_sec = ms/1000000;
    tv.tv_usec = ms%100000;
  }

  FD_ZERO(&mask);
  FD_SET(c->fd, &mask);
  
  ret = select(FD_SETSIZE, &mask, NULL, NULL, (ms >= 0) ? &tv : NULL);

#ifdef _REENTRANT
  _c_unlock(c);
#endif

  return ret;
}
